"""
Test Webdriver steps.
"""

from aloe.testing import FeatureTest

from aloe_webdriver.tests.base import feature, PAGES

# pylint:disable=line-too-long


class TestUtil(FeatureTest):
    """Test steps."""

    @feature()
    def test_I_should_see(self):
        """
Feature: I should see, I should not see
    Scenario: Everything fires up
        When I visit "{page}"
        Then I should see "Hello there!"
        And I should see 'Welcome "User"'
        And I should see a link to "Google" with the url "http://google.com/"
        And I should see a link with the url "http://google.com/"
        And I should not see "Bogeyman"
        """

        return dict(page=PAGES['basic_page'])

    @feature()
    def test_I_see_a_link(self):
        """
Feature: I should see a link
    Scenario: Everything fires up
        When I go to "{page}"
        Then  I should see a link to "Google" with the url "http://google.com/"
        And I see "Hello there!"
        """

        return dict(page=PAGES['basic_page'])

    @feature()
    def test_see_a_link_containing(self):
        """
Feature: I should see a link containing
    Scenario: Everything fires up
        When I go to "{page}"
        Then The browser's URL should contain "file://"
        And I should see a link that contains the text "Goo" and the url "http://google.com/"
        """

        return dict(page=PAGES['basic_page'])

    @feature()
    def test_basic_page_linking(self):
        """
Feature: Basic page linking
    Scenario: Follow links
        Given I go to "{link_page}"
        And I see "Page o link"
        When I click "Next Page"
        Then I should be at "{link_dest_page}"
        And The browser's URL should be "{link_dest_page}"
        And The browser's URL should not contain "http://"
        """

        return {
            'link_page': PAGES['link_page'],
            'link_dest_page': PAGES['link_dest']
        }

    @feature()
    def test_I_see_a_form(self):
        """
Feature: I should see a form
    Scenario: Everything fires up
        When I go to "{page}"
        Then I should see a form that goes to "basic_page.html"
        And the element with id of "somediv" contains "Hello"
        And the element with id of "somediv" does not contain "bye"
        """

        return dict(page=PAGES['basic_page'])

    @feature()
    def test_I_fill_in_a_form(self):
        """
Feature: I fill in a form
    Scenario: Everything fires up
        Given I go to "{page}"
        And I fill in "bio" with "everything awesome"
        And I fill in "Password: " with "neat"
        When I press "Submit!"
        Then The browser's URL should contain "bio=everything"
        """

        return dict(page=PAGES['basic_page'])

    @feature()
    def test_checkboxes_checked(self):
        """
Feature: Checkboxes checked
    Scenario: Everything fires up
        Given I go to "{page}"
        When I check "I have a bike"
        Then The "I have a bike" checkbox should be checked
        And The "I have a car" checkbox should not be checked
        """

        return dict(page=PAGES['basic_page'])

    @feature()
    def test_checkboxes_unchecked(self):
        """
Feature: Checkboxes unchecked
    Scenario: Everything fires up
        Given I go to "{page}"
        And I check "I have a bike"
        And The "I have a bike" checkbox should be checked
        When I uncheck "I have a bike"
        Then The "I have a bike" checkbox should not be checked
        """

        return dict(page=PAGES['basic_page'])

    @feature()
    def test_combo_boxes(self):
        """
Feature: Combo boxes
    Scenario: Everything fires up
        Given I go to "{page}"
        Then I should see option "Mercedes" in selector "car_choice"
        And I should see option "Volvo" in selector "car_choice"
        And I should not see option "Skoda" in selector "car_choice"
        When I select "Mercedes" from "car_choice"
        Then The "Mercedes" option from "car_choice" should be selected
        """

        return dict(page=PAGES['basic_page'])

    @feature(fails=True)
    def test_combo_boxes_fail(self):
        """
Feature: Combo boxes fail
    Scenario: Everything fires up
        Given I go to "{page}"
        Then I should not see option "Mercedes" in selector "car_choice"
        """

        return dict(page=PAGES['basic_page'])

    @feature()
    def test_multi_combo_boxes(self):
        '''
Feature: Multi-combo-boxes
    Scenario: Everything fires up
        Given I go to "{page}"
        When I select the following from "Favorite Colors:":
            """
            Blue
            Green
            """
        Then The following options from "Favorite Colors:" should be selected:
            """
            Blue
            Green
            """
        '''

        return dict(page=PAGES['basic_page'])

    @feature()
    def test_radio_buttons(self):
        """
Feature: Radio buttons
    Scenario: Everything fires up
        When I go to "{page}"
        And I choose "Male"
        Then The "Male" option should be chosen
        And The "Female" option should not be chosen
        """

        return dict(page=PAGES['basic_page'])

    @feature(fails=True)
    def test_hidden_text(self):
        """
Feature: Hidden text
    Scenario: Everything fires up
        When I go to "{page}"
        Then I should see an element with id of "bio_field"
        And I should see an element with id of "somediv" within 2 seconds
        And I should not see an element with id of "hidden_text"
        And I should see "Weeeee" within 1 second
        """

        return dict(page=PAGES['basic_page'])

    @feature(fails=True)
    def test_hidden_text_2(self):
        """
Feature: Hidden text 2
    Scenario: Everything fires up
        When I go to "{page}"
        Then I should see "Hello there" within 1 second
        And I should see an element with id of "oops_field" within 1 second
        And I should not see an element with id of "hidden_text"
        """

        return dict(page=PAGES['basic_page'])

    @feature()
    def test_alert_accept(self):
        """
Feature: test alert accept
    Scenario: alerts
        When I go to "{page}"
        Then I should see an alert with text "This is an alerting alert"
        When I accept the alert
        Then I should not see an alert
        And I should see "true"
        """

        return dict(page=PAGES['alert_page'])

    @feature()
    def test_alert_dismiss(self):
        """
Feature: test alert accept
    Scenario: alerts
        When I go to "{page}"
        Then I should see an alert with text "This is an alerting alert"
        When I dismiss the alert
        Then I should not see an alert
        And I should see "false"
        """

        return dict(page=PAGES['alert_page'])

    @feature()
    def test_tooltips(self):
        """
Feature: test tooltips
    Scenario: tooltips
        When I go to "{page}"
        Then I should see an element with tooltip "A tooltip"
        And I should not see an element with tooltip "Does not exist"
        And I should not see an element with tooltip "Hidden"
        When I click the element with tooltip "A tooltip"
        Then the browser's URL should contain "#anchor"
        """

        return dict(page=PAGES['tooltips'])

    @feature()
    def test_labels(self):
        """
Feature: test labels
    Scenario: basic page
        When I go to "{page}"
        And I click on label "Favorite Colors:"
        Then element with id "fav_colors" should be focused
        And element with id "bio_field" should not be focused
        """

        return dict(page=PAGES['basic_page'])

    @feature(fails=True)
    def test_labels_fail(self):
        """
Feature: test labels fail
    Scenario: basic page
        When I go to "{page}"
        And I click on label "Favorite Colors:"
        Then element with id "fav_colors" should not be focused
        """

        return dict(page=PAGES['basic_page'])

    @feature()
    def test_input_values(self):
        """
Feature: assert value
    Scenario: basic page
        When I go to "{page}"
        And I fill in "username" with "Danni"
        Then input "username" has value "Danni"
        """

        return dict(page=PAGES['basic_page'])

    @feature(fails=True)
    def test_input_values_fail(self):
        """
Feature: assert value
    Scenario: basic page
        When I go to "{page}"
        And I fill in "username" with "Danni"
        Then input "username" has value "Ricky"
        """

        return dict(page=PAGES['basic_page'])

    @feature()
    def test_date_input(self):
        """
Feature: assert value
    Scenario: basic page
        When I go to "{page}"
        And I fill in "dob" with "1900/01/01"
        Then input "dob" has value "1900/01/01"
        """

        return dict(page=PAGES['basic_page'])

    @feature()
    def test_page_title(self):
        """
Feature: assert value
    Scenario: basic page
        When I go to "{page}"
        Then the page title should be "A Basic Page"
        """

        return dict(page=PAGES['basic_page'])

    @feature()
    def test_submit_only(self):
        """
Feature: submit only form
    Scenario: basic page
        When I go to "{page}"
        And I submit the only form
        Then the browser's URL should contain "bio="
        And the browser's URL should contain "user="
        """

        return dict(page=PAGES['basic_page'])

    @feature()
    def test_submit_action(self):
        """
Feature: submit only form
    Scenario: basic page
        When I go to "{page}"
        And I submit the form with action "basic_page.html"
        Then the browser's URL should contain "bio="
        And the browser's URL should contain "user="
        """

        return dict(page=PAGES['basic_page'])

    @feature()
    def test_submit_id(self):
        """
Feature: submit only form
    Scenario: basic page
        When I go to "{page}"
        And I submit the form with id "the-form"
        Then the browser's URL should contain "bio="
        And the browser's URL should contain "user="
        """

        return dict(page=PAGES['basic_page'])



"""
Compatibility file to prevent API breakage
"""

from __future__ import absolute_import

import warnings

warnings.warn("This module is deprecated. Import aloe_webdriver directly.",
              DeprecationWarning)

# pylint:disable=wildcard-import, unused-import, redefined-builtin
from . import *



"""
Base functions for tests.
"""

import os
from functools import wraps

from aloe.testing import in_directory


def feature(fails=False):
    """
    Decorate a test method to test the feature contained in its docstring.

    Apply the context returned by the method to the feature.

    For example:
        @feature(failed=False)
        def test_some_feature(self):
            '''
            Feature: This name is returned
                Scenario: ...
                    When I {variable}
            '''

            return dict(variable=something)
    """

    def outer(func):
        """
        A decorator to run the function as the feature contained in docstring.
        """

        @wraps(func)
        @in_directory('tests')
        def inner(self):
            """Run the feature from docstring."""
            params = func(self)
            feature_string = func.__doc__.format(**params)

            result = self.run_feature_string(feature_string)

            if fails:
                self.assertFalse(result.success)
            else:
                self.assertTrue(result.success)

        return inner

    return outer


PAGES_DIR = os.path.join(os.path.dirname(__file__), 'html_pages')


PAGES = {}
for filename in os.listdir(PAGES_DIR):
    name = filename.split('.html')[0]
    PAGES[name] = 'file://%s' % os.path.join(PAGES_DIR, filename)



"""
Steps for selecting elements using CSS selectors.

Like with steps based on HTML id, these steps should be used cautiously to
avoid creating tests that do not describe the behaviours of your application.
See :ref:`good-bdd`.

.. note::

    Be aware these steps require jQuery_. If jQuery_ is not present it will be
    added (v1.10).

.. _jQuery: https://jquery.com/
"""

from functools import wraps
from time import sleep

from aloe import step
from aloe import world

from aloe_webdriver.util import (
    wait_for,
)

from nose.tools import (
    assert_equal,
    assert_false,
    assert_true,
)

from selenium.common.exceptions import WebDriverException

# pylint:disable=missing-docstring

# Pylint cannot infer the attributes on world
# pylint:disable=no-member


@wait_for
def wait_for_elem_selector(browser, sel):
    """Wait until the element specified by jQuery selector appears."""
    return find_elements_by_jquery(browser, sel)


def load_script(browser, url):
    """Ensure that JavaScript at a given URL is available to the browser."""
    if browser.current_url.startswith('file:'):
        url = 'https:' + url
    browser.execute_script("""
    var script_tag = document.createElement("script");
    script_tag.setAttribute("type", "text/javascript");
    script_tag.setAttribute("src", arguments[0]);
    document.getElementsByTagName("head")[0].appendChild(script_tag);
    """, url)

    sleep(1)


JQUERY = '//ajax.googleapis.com/ajax/libs/jquery/1.10.2/jquery.min.js'


def load_jquery(func):
    """
    A decorator to ensure a function is run with jQuery available.

    If an exception from a function indicates jQuery is missing, it is loaded
    and the function re-executed.

    The browser to load jQuery into must be the first argument of the function.
    """

    @wraps(func)
    def wrapped(browser, *args, **kwargs):
        """Run the function, loading jQuery if needed."""

        try:
            return func(browser, *args, **kwargs)
        except WebDriverException as ex:
            if ex.msg.startswith('$ is not defined'):
                load_script(browser, JQUERY)
                return func(browser, *args, **kwargs)
            else:
                raise

    return wrapped


@load_jquery
def find_elements_by_jquery(browser, selector):
    """Find HTML elements using jQuery-style selectors.

    Ensures that jQuery is available to the browser."""

    return browser.execute_script(
        """return ($ || jQuery)(arguments[0]).get();""", selector)


def find_element_by_jquery(browser, selector):
    """Find a single HTML element using jQuery-style selectors."""
    elements = find_elements_by_jquery(browser, selector)
    assert_true(len(elements) > 0)
    return elements[0]


@load_jquery
def find_parents_by_jquery(browser, selector):
    """Find HTML elements' parents using jQuery-style selectors.

    Ensures that jQuery is available to the browser."""

    return browser.execute_script(
        """return ($ || jQuery)(arguments[0]).parent().get();""", selector)


@step(r'There should be an element matching \$\("(.*?)"\)$')
def check_element_by_selector(self, selector):
    """Assert an element exists matching the given selector."""
    elems = find_elements_by_jquery(world.browser, selector)
    assert_true(elems)


@step(r'There should not be an element matching \$\("(.*?)"\)$')
def check_no_element_by_selector(self, selector):
    """Assert an element does not exist matching the given selector."""
    elems = find_elements_by_jquery(world.browser, selector)
    assert_false(elems)


@step(r'There should be an element matching \$\("(.*?)"\) '
      r'within (\d+) seconds?$')
def wait_for_element_by_selector(self, selector, seconds):
    """
    Assert an element exists matching the given selector within the given time
    period.
    """
    # pylint:disable=unexpected-keyword-arg
    # wait_for decorator parses the argument
    elems = wait_for_elem_selector(
        world.browser, selector, timeout=int(seconds))
    assert_true(elems)


@step(r'There should be exactly (\d+) elements matching \$\("(.*?)"\)$')
def count_elements_exactly_by_selector(self, number, selector):
    """
    Assert n elements exist matching the given selector.
    """
    elems = find_elements_by_jquery(world.browser, selector)
    assert_equal(len(elems), int(number))


@step(r'I fill in \$\("(.*?)"\) with "(.*?)"$')
def fill_in_by_selector(self, selector, value):
    """Fill in the form element matching the CSS selector."""
    elem = find_element_by_jquery(world.browser, selector)
    elem.clear()
    elem.send_keys(value)


@step(r'I submit \$\("(.*?)"\)')
def submit_by_selector(self, selector):
    """Submit the form matching the CSS selector."""
    elem = find_element_by_jquery(world.browser, selector)
    elem.submit()


@step(r'I check \$\("(.*?)"\)$')
def check_by_selector(self, selector):
    """Check the checkbox matching the CSS selector."""
    elem = find_element_by_jquery(world.browser, selector)
    if not elem.is_selected():
        elem.click()


@step(r'I click \$\("(.*?)"\)$')
def click_by_selector(self, selector):
    """Click the element matching the CSS selector."""
    # No need for separate button press step with selector style.
    elem = find_element_by_jquery(world.browser, selector)
    elem.click()


@step(r'I follow the link \$\("(.*?)"\)$')
def follow_link_by_selector(self, selector):
    """
    Navigate to the href of the element matching the CSS selector.

    N.B. this does not click the link, but changes the browser's URL.
    """
    elem = find_element_by_jquery(world.browser, selector)
    href = elem.get_attribute('href')
    world.browser.get(href)


@step(r'\$\("(.*?)"\) should be selected$')
def is_selected_by_selector(self, selector):
    """Assert the option matching the CSS selector is selected."""
    # No need for separate button press step with selector style.
    elem = find_element_by_jquery(world.browser, selector)
    assert_true(elem.is_selected())


@step(r'I select \$\("(.*?)"\)$')
def select_by_selector(self, selector):
    """Select the option matching the CSS selector."""
    option = find_element_by_jquery(world.browser, selector)
    selectors = find_parents_by_jquery(world.browser, selector)
    assert_true(len(selectors) > 0)
    selector = selectors[0]
    selector.click()
    sleep(0.3)
    option.click()
    assert_true(option.is_selected())



"""
Basic Selenium_ :class:`Webdriver` steps for Aloe_.
"""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import
# pylint:disable=redefined-builtin
from builtins import str
# pylint:enable=redefined-builtin

from aloe import step, world

from aloe_webdriver.util import (
    find_any_field,
    find_button,
    find_field,
    find_option,
    option_in_select,
    wait_for,
    string_literal,
)

from nose.tools import (
    assert_equal,
    assert_false,
    assert_in,
    assert_not_in,
    assert_true,
)

from selenium.webdriver.support.ui import Select
from selenium.webdriver.common.alert import Alert
from selenium.webdriver.common.keys import Keys
from selenium.common.exceptions import (
    NoSuchElementException,
    StaleElementReferenceException,
    NoAlertPresentException,
    WebDriverException)

# pylint:disable=missing-docstring

# Pylint cannot infer the attributes on world
# pylint:disable=no-member


def contains_content(browser, content):
    """
    Search for an element that contains the whole of the text we're looking
    for in it or its subelements, but whose children do NOT contain that
    text - otherwise matches <body> or <html> or other similarly useless
    things.
    """
    for elem in browser.find_elements_by_xpath(str(
            '//*[contains(normalize-space(.), {content}) '
            'and not(./*[contains(normalize-space(.), {content})])]'
            .format(content=string_literal(content)))):

        try:
            if elem.is_displayed():
                return True
        except StaleElementReferenceException:
            pass

    return False


@wait_for
def wait_for_elem(browser, xpath):
    return browser.find_elements_by_xpath(str(xpath))


@wait_for
def wait_for_content(browser, content):
    return contains_content(browser, content)

# Navigation ################################################################


@step('I visit "(.*?)"$')
@step('I go to "(.*?)"$')
def visit(self, url):
    """Navigate to the provided (fully qualified) URL."""
    world.browser.get(url)


@step('The browser\'s URL should be "([^"]*)"$')
@step('I should be at "([^"]*)"$')
def url_should_be(self, url):
    """Assert the absolute URL of the browser is as provided."""
    assert_equal(url, world.browser.current_url)


@step('''The browser's URL should contain "([^"]*)"$''')
def url_should_contain(self, url):
    """Assert the absolute URL of the browser contains the provided."""
    assert_in(url, world.browser.current_url)


@step('''The browser's URL should not contain "([^"]*)"$''')
def url_should_not_contain(self, url):
    """Assert the absolute URL of the browser does not contain the provided."""
    assert_not_in(url, world.browser.current_url)


@step(r'The page title should be "([^"]*)"')
@step(r"The page title should be '([^']*)'")
def page_title(self, title):
    """
    Assert the page title matches the given text.
    """
    assert_equal(world.browser.title, title)

# Links #####################################################################


@step('I click "([^"]*)"$')
@step("I click '([^']*)'$")
def click(self, name):
    """Click the link with the provided link text."""
    elem = world.browser.find_element_by_link_text(name)
    elem.click()


@step('I should see a link with the url "([^"]*)"$')
def should_see_link(self, link_url):
    """Assert a link with the provided URL is visible on the page."""
    assert_true(world.browser.
                find_element_by_xpath(str('//a[@href="%s"]' % link_url)))


@step('I should see a link to "([^"]*)" with the url "([^"]*)"$')
@step("I should see a link to '([^']*)' with the url '([^']*)'$")
def should_see_link_text(self, link_text, link_url):
    """Assert a link with the provided text points to the provided URL."""
    assert_true(world.browser.find_element_by_xpath(str(
        '//a[@href="%s"][./text()="%s"]' %
        (link_url, link_text))))


@step('I should see a link that contains the text "([^"]*)" and '
      'the url "([^"]*)"$')
@step("I should see a link that contains the text '([^']*)' and "
      "the url '([^']*)'$")
def should_include_link_text(self, link_text, link_url):
    """
    Assert a link containing the provided text points to the provided URL.
    """
    return world.browser.find_element_by_xpath(str(
        '//a[@href="%s"][contains(., %s)]' %
        (link_url, string_literal(link_text))))


# ID based checks ###########################################################


@step('The element with id of "([^"]*)" contains "([^"]*)"$')
@step("The element with id of '([^']*)' contains '([^']*)'$")
def element_contains(self, element_id, value):
    """
    Assert provided content is contained within an element found by ``id``.
    """
    return world.browser.find_element_by_xpath(str(
        'id("{id}")[contains(., "{value}")]'.format(
            id=element_id, value=value)))


@step('The element with id of "([^"]*)" does not contain "([^"]*)"$')
@step("The element with id of '([^']*)' does not contain '([^']*)'$")
def element_not_contains(self, element_id, value):
    """
    Assert provided content is not contained within an element found by ``id``.
    """
    elem = world.browser.find_elements_by_xpath(str(
        'id("{id}")[contains(., "{value}")]'.format(
            id=element_id, value=value)))
    assert_false(elem)


@wait_for
def wait_for_visible_elem(browser, xpath):
    elem = browser.find_elements_by_xpath(str(xpath))
    if not elem:
        return False
    return elem[0].is_displayed()


@step(r'I should see an element with id of "([^"]*)" within (\d+) seconds?$')
def should_see_id_in_seconds(self, element_id, timeout):
    """
    Assert an element with the given ``id`` is visible within n seconds.
    """
    # pylint:disable=unexpected-keyword-arg
    # wait_for decorator parses the argument
    elem = wait_for_visible_elem(world.browser, 'id("%s")' % element_id,
                                 timeout=int(timeout))
    assert_true(elem)


@step('I should see an element with id of "([^"]*)"$')
def should_see_id(self, element_id):
    """
    Assert an element with the given ``id`` is visible.
    """
    elem = world.browser.find_element_by_xpath(str('id("%s")' % element_id))
    assert_true(elem.is_displayed())


@step('I should not see an element with id of "([^"]*)"$')
def should_not_see_id(self, element_id):
    """
    Assert an element with the given ``id`` is not visible.
    """
    try:
        elem = world.browser.find_element_by_xpath(str('id("%s")' %
                                                       element_id))
        assert_false(elem.is_displayed())
    except NoSuchElementException:
        pass


@step(r'Element with id "([^"]*)" should be focused')
def element_focused(self, id_):
    """
    Assert the element is focused.
    """

    elem = world.browser.find_element_by_xpath(
        str('id("{id}")'.format(id=id_)))
    focused = world.browser.switch_to.active_element

    assert_true(elem == focused)


@step(r'Element with id "([^"]*)" should not be focused')
def element_not_focused(self, id_):
    """
    Assert the element is not focused.
    """

    elem = world.browser.find_element_by_xpath(
        str('id("{id}")'.format(id=id_)))
    focused = world.browser.switch_to.active_element

    # Elements don't have __ne__ defined, cannot test for inequality
    assert_false(elem == focused)


# Text ######################################################################


@step(r'I should see "([^"]+)" within (\d+) seconds?$')
@step(r"I should see '([^']+)' within (\d+) seconds?$")
def should_see_in_seconds(self, text, timeout):
    """
    Assert provided text is visible within n seconds.

    Be aware this text could be anywhere on the screen. Also be aware that
    it might cross several HTML nodes. No determination is made between
    block and inline nodes. Whitespace can be affected.
    """
    # pylint:disable=unexpected-keyword-arg
    # wait_for decorator parses the argument
    assert_true(wait_for_content(world.browser, text, timeout=int(timeout)))


@step('I should see "([^"]+)"$')
@step("I should see '([^']+)'$")
@step('I see "([^"]+)"$')
@step("I see '([^']+)'$")
def should_see(self, text):
    """
    Assert provided text is visible.

    Be aware this text could be anywhere on the screen. Also be aware that
    it might cross several HTML nodes. No determination is made between
    block and inline nodes. Whitespace can be affected.
    """
    assert_true(contains_content(world.browser, text))


@step('I should not see "([^"]+)"$')
@step("I should not see '([^']+)'$")
def should_not_see(self, text):
    """
    Assert provided text is not visible.

    Be aware that because of the caveats of the positive case, the text MAY
    be on the screen in a slightly different form.
    """
    assert_false(contains_content(world.browser, text))


# Forms #####################################################################


@step('I should see a form that goes to "([^"]*)"$')
def see_form(self, url):
    """
    Assert the existence of a HTML form that submits to the given URL.
    """
    return world.browser.find_element_by_xpath(str('//form[@action="%s"]' %
                                                   url))


DATE_FIELDS = (
    'datetime',
    'datetime-local',
    'date',
)


TEXT_FIELDS = (
    'text',
    'textarea',
    'password',
    'month',
    'time',
    'week',
    'number',
    'range',
    'email',
    'url',
    'tel',
    'color',
)


@step('I fill in "([^"]*)" with "([^"]*)"$')
@step("I fill in '([^']*)' with '([^']*)'$")
def fill_in_textfield(self, field_name, value):
    """
    Fill in the HTML input with given label (recommended), name or id with
    the given text.

    Supported input types are text, textarea, password, month, time, week,
    number, range, email, url, tel and color.
    """
    date_field = find_any_field(world.browser,
                                DATE_FIELDS,
                                field_name)
    if date_field:
        field = date_field
    else:
        field = find_any_field(world.browser,
                               TEXT_FIELDS,
                               field_name)

    assert_true(field,
                'Can not find a field named "%s"' % field_name)
    if date_field:
        field.send_keys(Keys.DELETE)
    else:
        field.clear()
    field.send_keys(value)


@step('I press "([^"]*)"$')
@step("I press '([^']*)'$")
def press_button(self, value):
    """
    Click the button with the given label.
    """
    button = find_button(world.browser, value)
    button.click()


@step('I click on label "([^"]*)"')
@step("I click on label '([^']*)'")
def click_on_label(self, label):
    """
    Click on the given label.

    On a correctly set up form this will highlight the appropriate field.
    """

    elem = world.browser.find_element_by_xpath(str(
        '//label[normalize-space(text())=%s]' % string_literal(label)))
    elem.click()


@step(r'Input "([^"]*)" (?:has|should have) value "([^"]*)"')
@step(r"Input '([^']*)' (?:has|should have) value '([^']*)'")
def input_has_value(self, field_name, value):
    """
    Assert the form input with label (recommended), name or id has given value.
    """
    text_field = find_any_field(world.browser,
                                DATE_FIELDS + TEXT_FIELDS,
                                field_name)
    assert_false(text_field is False,
                 'Can not find a field named "%s"' % field_name)
    assert_equal(text_field.get_attribute('value'), value)


@step(r'I submit the only form')
def submit_the_only_form(self):
    """
    Look for a form on the page and submit it.

    Asserts if more than one form exists.
    """
    form = world.browser.find_element_by_xpath(str('//form'))
    form.submit()


@step(r'I submit the form with id "([^"]*)"')
def submit_form_id(self, id_):
    """
    Submit the form with given id (used to disambiguate between multiple
    forms).
    """
    form = world.browser.find_element_by_xpath(
        str('id("{id}")'.format(id=id_)))
    form.submit()


@step(r'I submit the form with action "([^"]*)"')
def submit_form_action(self, url):
    """
    Submit the form with the given action URL (i.e. the form that submits to
    ``/post/my/data``).
    """
    form = world.browser.find_element_by_xpath(str('//form[@action="%s"]' %
                                                   url))
    form.submit()


# Checkboxes ################################################################


@step('I check "([^"]*)"$')
@step("I check '([^']*)'$")
def check_checkbox(self, value):
    """Check the checkbox with label (recommended), name or id."""
    check_box = find_field(world.browser, 'checkbox', value)
    if not check_box.is_selected():
        check_box.click()


@step('I uncheck "([^"]*)"$')
@step("I uncheck '([^']*)'$")
def uncheck_checkbox(self, value):
    """Uncheck the checkbox with label (recommended), name or id."""
    check_box = find_field(world.browser, 'checkbox', value)
    if check_box.is_selected():
        check_box.click()


@step('The "([^"]*)" checkbox should be checked$')
@step("The '([^']*)' checkbox should be checked$")
def assert_checked_checkbox(self, value):
    """Assert the checkbox with label (recommended), name or id is checked."""
    check_box = find_field(world.browser, 'checkbox', value)
    assert_true(check_box.is_selected())


@step('The "([^"]*)" checkbox should not be checked$')
@step("The '([^']*)' checkbox should not be checked$")
def assert_not_checked_checkbox(self, value):
    """
    Assert the checkbox with label (recommended), name or id is not checked.
    """
    check_box = find_field(world.browser, 'checkbox', value)
    assert_false(check_box.is_selected())


# Selects ###################################################################


@step('I select "([^"]*)" from "([^"]*)"$')
@step("I select '([^']*)' from '([^']*)'$")
def select_single_item(self, option_name, select_name):
    """
    Select the named option from select with label (recommended), name or id.
    """
    option_box = find_option(world.browser, select_name, option_name)
    option_box.click()


@step('I select the following from "([^"]*?)":?$')
@step("I select the following from '([^']*?)':?$")
def select_multi_items(self, select_name):
    """
    Select multiple options from select with label (recommended), name, or
    id. Pass a multiline string of options. e.g.

    .. code-block:: gherkin

        When I select the following from "Contact Methods":
            \"\"\"
            Email
            Phone
            Fax
            \"\"\"
    """
    # Ensure only the options selected are actually selected
    option_names = self.multiline.split('\n')
    select_box = find_field(world.browser, 'select', select_name)

    select = Select(select_box)
    select.deselect_all()

    for option in option_names:
        try:
            select.select_by_value(option)
        except NoSuchElementException:
            select.select_by_visible_text(option)


@step('The "([^"]*)" option from "([^"]*)" should be selected$')
@step("The '([^']*)' option from '([^']*)' should be selected$")
def assert_single_selected(self, option_name, select_name):
    """
    Assert the given option is selected from the select with label
    (recommended), name or id.

    If multiple selections are supported other options may be selected.
    """
    option = find_option(world.browser, select_name, option_name)
    assert_true(option.is_selected())


@step('The following options from "([^"]*?)" should be selected:?$')
@step("The following options from '([^']*?)' should be selected:?$")
def assert_multi_selected(self, select_name):
    # Ensure its not selected unless its one of our options
    option_names = self.multiline.split('\n')
    select_box = find_field(world.browser, 'select', select_name)
    option_elems = select_box.find_elements_by_xpath(str('./option'))
    for option in option_elems:
        if option.get_attribute('id') in option_names or \
           option.get_attribute('name') in option_names or \
           option.get_attribute('value') in option_names or \
           option.text in option_names:
            assert_true(option.is_selected())
        else:
            assert_false(option.is_selected())


@step(r'I should see option "([^"]*)" in selector "([^"]*)"')
@step(r"I should see option '([^']*)' in selector '([^']*)'")
def select_contains(self, option, id_):
    """Assert the select contains the given option."""
    assert_true(option_in_select(world.browser, id_, option) is not None)


@step(r'I should not see option "([^"]*)" in selector "([^"]*)"')
@step(r"I should not see option '([^']*)' in selector '([^']*)'")
def select_does_not_contain(self, option, id_):
    """Assert the select does not contain the given option."""
    assert_true(option_in_select(world.browser, id_, option) is None)


# Radios ####################################################################


@step('I choose "([^"]*)"$')
@step("I choose '([^']*)'$")
def choose_radio(self, value):
    """
    Click (and choose) the radio button with the given label (recommended),
    name or id.
    """
    box = find_field(world.browser, 'radio', value)
    box.click()


@step('The "([^"]*)" option should be chosen$')
@step("The '([^']*)' option should be chosen$")
def assert_radio_selected(self, value):
    """
    Assert the radio button with the given label (recommended), name or id is
    chosen.
    """
    box = find_field(world.browser, 'radio', value)
    assert_true(box.is_selected())


@step('The "([^"]*)" option should not be chosen$')
@step("The '([^']*)' option should not be chosen$")
def assert_radio_not_selected(self, value):
    """
    Assert the radio button with the given label (recommended), name or id is
    not chosen.
    """
    box = find_field(world.browser, 'radio', value)
    assert_false(box.is_selected())


# Alerts ####################################################################


@step('I accept the alert')
def accept_alert(self):
    """
    Accept the alert.
    """

    try:
        alert = Alert(world.browser)
        alert.accept()
    except WebDriverException:
        # PhantomJS is kinda poor
        pass


@step('I dismiss the alert')
def dismiss_alert(self):
    """
    Dismiss the alert.
    """

    try:
        alert = Alert(world.browser)
        alert.dismiss()
    except WebDriverException:
        # PhantomJS is kinda poor
        pass


@step(r'I should see an alert with text "([^"]*)"')
@step(r"I should see an alert with text '([^']*)'")
def check_alert(self, text):
    """
    Assert an alert is showing with the given text.
    """

    try:
        alert = Alert(world.browser)
        assert_equal(alert.text, text)
    except WebDriverException:
        # PhantomJS is kinda poor
        pass


@step('I should not see an alert')
def check_no_alert(self):
    """
    Assert there is no alert.
    """

    try:
        alert = Alert(world.browser)
        raise AssertionError("Should not see an alert. Alert '%s' shown." %
                             alert.text)
    except NoAlertPresentException:
        pass

# Tooltips ##################################################################


@step(r'I should see an element with tooltip "([^"]*)"')
@step(r"I should see an element with tooltip '([^']*)'")
def see_tooltip(self, tooltip):
    """
    Assert an element with the given tooltip (title) is visible.

    N.B. tooltip may not be visible.
    """
    elem = world.browser.find_elements_by_xpath(str(
        '//*[@title=%(tooltip)s or @data-original-title=%(tooltip)s]' %
        dict(tooltip=string_literal(tooltip))))
    elem = [e for e in elem if e.is_displayed()]
    assert_true(elem)


@step(r'I should not see an element with tooltip "([^"]*)"')
@step(r"I should not see an element with tooltip '([^']*)'")
def no_see_tooltip(self, tooltip):
    """
    Assert an element with the given tooltip (title) is not visible.
    """
    elem = world.browser.find_elements_by_xpath(str(
        '//*[@title=%(tooltip)s or @data-original-title=%(tooltip)s]' %
        dict(tooltip=string_literal(tooltip))))
    elem = [e for e in elem if e.is_displayed()]
    assert_false(elem)


@step(r'I (?:click|press) the element with tooltip "([^"]*)"')
@step(r"I (?:click|press) the element with tooltip '([^']*)'")
def press_by_tooltip(self, tooltip):
    """
    Click on a HTML element with a given tooltip.

    This is very useful if you're clicking on icon buttons, etc.
    """
    for button in world.browser.find_elements_by_xpath(str(
            '//*[@title="%(tooltip)s" or @data-original-title="%(tooltip)s"]'
            % dict(tooltip=tooltip)
    )):
        try:
            button.click()
            break
        except:  # pylint:disable=bare-except
            pass
    else:
        raise AssertionError("No button with tooltip '{0}' found"
                             .format(tooltip))

# Frames ####################################################################


@step(r'I switch to the frame with id "([^"]*)"')
def switch_to_frame(self, frame):
    """Swap Selenium's context to the given frame or iframe."""
    elem = world.browser.find_element_by_id(frame)
    world.browser.switch_to.frame(elem)


@step(r'I switch back to the main view')
def switch_to_main(self):
    """Swap Selenium's context back to the main window."""
    world.browser.switch_to.default_content()



"""
Steps for testing.
"""

from contextlib import contextmanager
try:
    reload
except NameError:
    # pylint:disable=no-name-in-module,redefined-builtin
    from importlib import reload
    # pylint:enable=no-name-in-module,redefined-builtin

from selenium import webdriver

from aloe import around, before, world

import aloe_webdriver
import aloe_webdriver.css

# This module is reloaded during testing in order to re-register the steps and
# callbacks. Make sure the modules where the steps are defined are, too.
reload(aloe_webdriver)
reload(aloe_webdriver.css)


@around.all
@contextmanager
def with_browser():
    """Start a browser for the tests."""
    world.browser = webdriver.Firefox()
    world.browser.get('')
    yield
    world.browser.quit()
    delattr(world, 'browser')


@before.each_feature
def reset_page(feature):
    """Reset the browser before each feature."""
    world.browser.get('')



"""Steps and utility functions for taking screenshots."""

import time
import uuid

from aloe import (
    after,
    step,
    world,
)
import os.path
import json

# Pylint cannot infer the attributes on world
# pylint:disable=no-member


def set_save_directory(base, source):
    """Sets the root save directory for saving screenshots.

    Screenshots will be saved in subdirectories under this directory by
    browser window size. """
    root = os.path.join(base, source)
    if not os.path.isdir(root):
        os.makedirs(root)

    world.screenshot_root = root


def resolution_path():
    """Make up a directory path based on browser resolution."""
    window_size = world.browser.get_window_size()
    return os.path.join(
        world.screenshot_root,
        '{}x{}'.format(window_size['width'], window_size['height']),
    )


@step(r'I capture a screenshot$')
def capture_screenshot(self):
    """Capture a screenshot and save the name it's written as."""
    feature = self.scenario.feature
    self.shot_name = '{}.png'.format(uuid.uuid4())
    if getattr(feature, 'dir_path', None) is None:
        feature.dir_path = resolution_path()
    if not os.path.isdir(feature.dir_path):
        os.makedirs(feature.dir_path)
    filename = os.path.join(
        feature.dir_path,
        self.shot_name,
    )
    world.browser.get_screenshot_as_file(filename)


@step(r'I capture a screenshot after (\d+) seconds?$')
def capture_screenshot_delay(self, delay):
    """I capture a screenshot after a delay."""
    time.sleep(delay)
    capture_screenshot(self)


@after.each_feature
def record_run_feature_report(feature):
    """Write a report of all screenshots corresponding to the features."""
    if getattr(feature, 'dir_path', None) is None:
        return
    feature_name_json = '{}.json'.format(os.path.splitext(
        os.path.basename(feature.described_at.file)
    )[0])
    report = {}
    for scenario in feature.scenarios:
        scenario_report = []
        for self in scenario.steps:
            shot_name = getattr(self, 'shot_name', None)
            if shot_name is not None:
                scenario_report.append(shot_name)
        if scenario_report:
            report[scenario.name] = scenario_report

    if report:
        with open(os.path.join(feature.dir_path, feature_name_json), 'w') \
                as report_file:
            json.dump(report, report_file)



# -*- coding: utf-8 -*-
"""Test step functions directly."""

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import unittest

from selenium import webdriver

from aloe import world
from aloe_webdriver.tests.base import PAGES
from aloe_webdriver.util import (
    find_button,
    find_field,
    find_field_by_id,
    find_field_by_label,
    find_field_by_name,
    option_in_select,
    wait_for,
)

# pylint:disable=missing-docstring


class TestUtil(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        world.browser = webdriver.Firefox()

    def setUp(self):
        world.browser.get(PAGES['basic_page'])

    @classmethod
    def tearDownClass(cls):
        world.browser.quit()
        delattr(world, 'browser')

    def test_find_by_id(self):
        assert find_field_by_id(world.browser, 'password', 'pass')

    def test_find_by_name(self):
        assert find_field_by_name(world.browser, 'submit', 'submit')
        assert find_field_by_name(world.browser, 'select', 'car_choice')
        assert find_field_by_name(world.browser, 'textarea', 'bio')

    def test_find_by_label(self):
        field = find_field_by_label(world.browser, 'text', 'Username:')
        self.assertEqual(field.get_attribute('id'), 'username')

    def test_no_label(self):
        assert not find_field_by_label(world.browser, 'text', 'NoSuchLabel')

    def test_find_field(self):
        assert find_field(world.browser, 'text', 'username')
        assert find_field(world.browser, 'text', 'Username:')
        assert find_field(world.browser, 'text', 'user')
        assert find_field(world.browser, 'text', 'ชื่อ:')

    def test_find_button(self):
        assert find_button(world.browser, 'submit')
        assert find_button(world.browser, 'Submit!')
        assert find_button(world.browser, 'submit_tentative')
        assert find_button(world.browser, 'Submit as tentative')
        assert find_button(world.browser, 'ส่งฟอร์ม')

    def test_option_in_select(self):
        assert option_in_select(world.browser, 'Favorite Colors:', 'Blue')
        assert option_in_select(world.browser, 'Favorite Colors:', 'ฟ้า')

    def test_wait_for(self):
        counter = [0]

        @wait_for
        def lazy_function(i):
            counter[0] += 1
            return counter[0] > i

        # pylint:disable=unexpected-keyword-arg
        # wait_for decorator parses the argument
        assert not lazy_function(10, timeout=1)
        assert lazy_function(5, timeout=1)



"""Test CSS selector steps."""

from aloe.testing import FeatureTest

from aloe_webdriver.tests.base import feature, PAGES


class TestCSS(FeatureTest):
    """Test CSS steps."""

    @feature()
    def test_css_match(self):
        # pylint:disable=line-too-long
        """
Feature: Wait and match CSS
    Scenario: Everything fires up
        When I go to "{page}"
        Then There should be an element matching $("textarea[name='bio']") within 1 second
        """
        # pylint:enable=line-too-long

        return dict(page=PAGES['basic_page'])

    @feature()
    def test_forms(self):
        """
Feature: CSS-based formstuff
    Scenario: Everything fires up
        When I go to "{page}"
        Then I fill in $("input[name='user']") with "A test string"
        And I check $("input[value='Bike']")
        """

        return dict(page=PAGES['basic_page'])
